/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright (C) 1998-2001  Gerwin Klein <lsf@jflex.de>                    *
 * All rights reserved.                                                    *
 *                                                                         *
 * 2002 - Modified     <a href="mailto:roni.hursti@luukku.com">            *
 * Roni Hursti</a>                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License. See the file      *
 * COPYRIGHT for more information.                                         *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                 *
 *                                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

import java.io.FileReader;

/**
 * Simple test driver for the static analysis for Java-class.
 * Reads arguments from command line
 */
public final class JavaParser {

  /** Options constants */
  private static final String NO_MEMBER = "-no_members";
  private static final String DEBUG = "-debug";
  private static final String VERBOSE = "-v";
  private static final String STRICT = "-all_members";
  private static final String DEF = "-all_constructors";
  private static final String NO_DEF = "-no_constructors";
  private static final String NO_METHOD = "-no_method_parameters";
  private static final String LINE = "-";

  /* Default constructor */
  private JavaParser() { }

  /* Print usage information */
  private static void printUsage() {
    System.out.println("Usage: " +" [OPTIONS] FILENAME\n\nwhere FILENAME is Java source file(s)"+
                       "\n\n[OPTIONS]\n\n"+
                       "\n"+DEBUG+"\t\tPrint debugging information"+
                       "\n"+VERBOSE+"\t\tPrint useful information\n"+
                       NO_MEMBER+"\tDo not check member variables"+
                       "\n"+STRICT+"\tCheck all member variables\n"+
                       "\t\t(Only private member variables will be checked by default)"+
                       "\n"+DEF+"\n\t\tCheck also that every inner\n"+
                       "\t\tclass (not anonymous) has default constructor"+
                       "\n"+NO_DEF+"\n\t\tDo not check default constructors"+
		       "\n"+NO_METHOD+"\n\t\tDo not check method parameters");
  }

  /** Main program
   * @param argv command line arguments
   */
  public static void main(String argv[]) {
    boolean member = true;
    boolean debug = false;
    boolean verbose = false;
    boolean strict = true;
    boolean construct = false;
    boolean no_construct = false;
    boolean no_method = true;

    if (argv == null || argv.length == 0) {
      printUsage();
      System.exit(1);
    }

    for (int i = 0; i < argv.length; i++) {
      boolean found = false;
      
      if (argv[i].startsWith(NO_MEMBER)) {
        argv[i] = null;
        member = false; found = true;
      }
      else if (argv[i].startsWith(DEBUG)) {
        argv[i] = null;
        debug = true; found = true;
      }
      else if (argv[i].startsWith(VERBOSE)) {
        argv[i] = null;
        verbose = true; found = true;
      }
      else if (argv[i].startsWith(STRICT)) {
        argv[i] = null;
        strict = false; found = true;
      }
      else if (argv[i].startsWith(DEF)) {
        argv[i] = null;
        construct = true; found = true;
      }
      else if (argv[i].startsWith(NO_DEF)) {
        argv[i] = null;
        no_construct = true; found = true;
      }
      else if (argv[i].startsWith(NO_METHOD)) {
        argv[i] = null;
        no_method = false; found = true;
      }
      // If option not found
      else if (argv[i].startsWith(LINE)) {
        argv[i] = null;
        printUsage();
        System.exit(1);
      }
        
    }
    
    try {
      for (int i = 0; i < argv.length; i++) {
        if (argv[i] != null) {
          OwnScanner s = new OwnScanner(new UnicodeEscapes(new FileReader(argv[i])));
          if (verbose) System.out.println("Doing static analysis for classname: " +argv[i]+"\n");
          parser p = new parser(s, argv[i], member, debug, strict, construct, no_method, no_construct);
          p.parse();
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
      System.exit(1);
    }
  }
}
