/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright (C) 2002  Roni Hursti <ronih@luukku.com>                      *
 * All rights reserved.                                                    *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License. See the file      *
 * COPYRIGHT for more information.                                         *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                 *
 *                                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

import java.util.HashMap;
import java.util.ArrayList;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * Check modifier occurences <code>VariableChecker</code>
 *
 * @author     <a href="mailto:roni.hursti@luukku.com">Roni Hursti</a>
 * @version    1.0
 * @see        StringTokenizer
 * @see        HashMap
 * @see        ArrayList
 * @see        Set
 * @since      jdk1.2
 */
public final class VariableChecker {

  /* This boolean variable is for debugging <code>printBoolean</code>
   * Setting the variable TRUE causes everything to be printed to stdout
   */
  private boolean printBoolean;

  /* This variable holds all modifiers (HashMap) in every blocks
   * <code>arraylist</code>
   * @see hashmap for modifiers in one block
   */
  private final ArrayList arrayList;

  /* This variable object holds the array list for temporary
   * use
   */
  private final ArrayList tempArrayList;

  /* This variable holds every modifiers in one block in
   * <code>VariableMap</code>
   */
  private VariableMap variableMap;

  /* This variable holds the information of the type of the modifier
   * (for example VARIABLE, IMPORT) <code>VARIABLE</code>
   */
  private final String VARIABLE;

  /* This variable holds the information of the previous line's
   * line number information
   */
  private int oldLineNumber;
  
  /* This is a boolean variable for holding information that modifier
   * is a member variable
   */
  private final boolean isMemberVariable;
    
  /* This variable holds the arraylist's index 
   */
  private int arrayListIndex;

  /* A constant for Method parameter variable
   */
  private static final String METHOD_VARIABLE = "Method parameter variable ";
  
  /**
   * Default constructor
   */
  public VariableChecker() {
    this.printBoolean = false;
    arrayList = new ArrayList();
    tempArrayList = new ArrayList();
    variableMap = new VariableMap();
    this.VARIABLE = "Variable ";
    this.isMemberVariable = false; this.arrayListIndex = 0;
  }
  
  /**
   * A constructor
   * @param s <code>String</code> for modifier printout
   * @param isMemberVariable <code>boolean</code> for member
   * variable information
   */
  public VariableChecker(final String s, final boolean isMemberVariable) {
    this.printBoolean = false;
    if (isEmpty(s))
      this.VARIABLE = "Variable ";
    else
      this.VARIABLE = s + " ";
    arrayList = new ArrayList();
    tempArrayList = new ArrayList();
    variableMap = new VariableMap();
    this.isMemberVariable = isMemberVariable; this.arrayListIndex = 0;
  }

  /** Set class name information
   * @param className value of the class name
   */
  public void setClassName(final String className) {
    variableMap.setClassName(className);
  }
  /** Get class name information
   * @return value of the class name 
   */
  public String getClassName() {
    return this.variableMap.getClassName();
  }

  /** Set debugging on
   */
  public void setPrintDebugging() {
    this.printBoolean = true;
  }

  /**
   * Remove the whole variableMap from arraylist
   */
  public void remove() {
    int size;

    if ((size = arrayList.size()) != 0)
      arrayList.remove(size - 1);
      
    if (printBoolean) {
      System.out.println("list after remove");
      printDebug();
    }
  }

  /** Check if there is an open method
   * @return true if method is still open
   */
  public boolean isMethodOpen() {
    return tempArrayList.size() > 0;
  }
  
  /** Return hashMap copy
   * @return hashmap
   */
  public HashMap returnHashMap() {
    if (printBoolean) {
      System.out.println("Returning hashmap");
      printHashMap();
    }
    return ((HashMap)this.variableMap.hashMap.clone());
  }
  
  /**
   * Check if the parameter string is empty. Return true if
   * it is empty and otherwise return false
   * @param s <code>String</code> input parameter for checking empty
   */
  public static boolean isEmpty(final String s) {
    return s == null || s.length() == 0;
  }

  /**
   * Print member HashMap table
   */
  public void printArrayList() {
    System.out.println("printing arraylist begins...");
    System.out.println("className: " +this.variableMap.getClassName());
    printHashMap();
    System.out.println("");
    for (int i = this.arrayList.size(); --i >= 0;) {
      VariableMap vm = (VariableMap)this.arrayList.get(i);
      System.out.println("lineNumber: " +vm.getLineNumber());
      System.out.println("className: " +vm.getClassName());
      System.out.println("hasClassStopped: " +vm.hasClassStopped());
      System.out.println("index: " +i);
      printHashMap
        (vm.hashMap);
    }
    System.out.println("printing arraylist ends...");
  }

  /**
   * Print member HashMap table
   * @param arrayList the arraylist which is printed out
   */
  public void printArrayList(final ArrayList arrayList) {
    System.out.println("printing arraylist begins...");
    for (int i = arrayList.size(); --i >= 0;) {
      VariableMap vm = (VariableMap)arrayList.get(i);
      System.out.println("lineNumber: " +vm.getLineNumber());
      System.out.println("className: " +vm.getClassName());
      System.out.println("hasClassStopped: " +vm.hasClassStopped());
      System.out.println("index: " +i);
      printHashMap
        (vm.hashMap);
    }
    System.out.println("printing arraylist ends...");
  }
    
  /**
   * Print hashMap information
   */
  public void printHashMap() {
    Set entry = this.variableMap.hashMap.entrySet();
    Object[] tableEntry = entry.toArray();

    System.out.println("Hashmap's size is " +tableEntry.length);
    
    for (int j = tableEntry.length; --j >= 0;) {
      String key = getKeyFromEntry(tableEntry[j]);
      VariableOccurence vo = 
        (VariableOccurence)this.variableMap.hashMap.get(key);
      System.out.println
        (key + " " + vo.getOccurences() + " declarated: " +vo.getDeclarated());
    }
  }

  /**
   * Print parameter's HashMap table
   * @param hashMap <code>HashMap</code> input parameter for printing
   * hashmap
   */
  public static void printHashMap(final HashMap hashMap) {
    if (hashMap == null || hashMap.size() == 0)
      return;
    Set entry = hashMap.entrySet();
    Object[] tableEntry = entry.toArray();

    for (int j = tableEntry.length; --j >= 0;) {
      String key = getKeyFromEntry(tableEntry[j]);
      VariableOccurence vo = 
        (VariableOccurence)hashMap.get(key);
      if (key != null)
        System.out.println
          (key + " " + vo.getOccurences
           () + " declarated: " +vo.getDeclarated());
    }
  }

  /** Store variable in tempArrayList
   * if variable exists in tempArrayList then return true
   * @param variable which is stored in tempArrayList if
   * exists
   * @return true if variable exists in tempArrayList
   */
  public boolean storeIfInTempList(final String variable) {
    int size = tempArrayList.size();
    
    if (size < 1)
      return false;
    
    for (int i = size; --i >= 0;) {
      VariableMap vm = (VariableMap)tempArrayList.get(i);
      VariableOccurence vo;
      if ((vo = (VariableOccurence)vm.hashMap.get(variable)) != null) {
        vm.hashMap.put
          (variable, new VariableOccurence
            (vo.getOccurences() + 1, vo.getDeclarated()));
        return true;
      }
    }

    return false;
  }

  /**
   * Add variable occurence to the hashmap
   * @param variable <code>String</code> which we add to the hashMap
   * @param variableChecker <code>VariableChecker</code> variableChecker object
   * (for member variables)
   * @param isMemberReference <code>boolean</code> is member reference
   */
  public void
  addVariable(final String variable, final VariableChecker variableChecker,
              final boolean isMemberReference) {

    VariableOccurence vo;

    if (isEmpty(variable))
      return;

    // if local variable exists
    // increase variable occurence count by 1
    if (!isMemberReference) {
      if ((vo = get(variable)) != null) {
        if (printBoolean) System.out.println("adding variable " +variable);
        this.variableMap.hashMap.put
          (variable, new VariableOccurence
            (vo.getOccurences() + 1, vo.getDeclarated()));
        return;
      } else
        if (storeIfInTempList(variable))
          return;
    }

    vo = (VariableOccurence)variableChecker.variableMap.hashMap.get
      (variable);
    if (vo != null) {
      if (printBoolean) 
        System.out.println
          ("Member reference "+variable+" already declared");
      variableChecker.variableMap.hashMap.put
        (variable, 
         new VariableOccurence(vo.getOccurences()+1, 
                               vo.getDeclarated()));
    }
    else {
      if (printBoolean) 
        System.out.println
          ("Member reference "+variable+" not declared yet");
      variableChecker.variableMap.hashMap.put
        (variable, new VariableOccurence
          (0, false));
    }
  }

  /**
   * Add variable occurence to the hashmap
   * @param variable <code>String</code> which we add to the hashMap
   * @param className <code>String</code>Class name for variable
   */
  public void
  addVariable(final String variable, 
              final String className) {

    if (isEmpty(variable))
      return;

    for (int i = this.arrayList.size(); --i >= 0;) {
      VariableOccurence vo;
      VariableMap vm = (VariableMap)this.arrayList.get(i);
      String lclassName;
      
      // If className found
      if ((lclassName = vm.getClassName()) != null
	  && lclassName.equals(className)) {

        // if member variable exists
        if ((vo = (VariableOccurence)vm.hashMap.get
	     (variable)) != null) {
          if (printBoolean) 
            System.out.println
              ("adding member reference outer variable " +variable);
          vm.hashMap.put(variable, new VariableOccurence
            (vo.getOccurences() + 1, 
	     vo.getDeclarated(), new VariableType(vo.getLineNumber())));
        }
        // Assume that member variable's reference occurs
	// before it's declaration

        else {
          if (printBoolean)
            System.out.println
              ("adding member variable reference " +variable);
          vm.hashMap.put(variable, new VariableOccurence(0, false));
        }

        this.arrayList.set(i, vm.clone());
        return;
      }
    }
  }

  /**
   * Add variable occurence to the hashmap
   * @param variable <code>String</code> which we add to the hashMap
   */
  public void addVariable(final String variable) {
    VariableOccurence intV;
    
    if (isEmpty(variable))
      return;

    // increase variable occurence count by 1
    if ((intV = get(variable)) != null) {
      if (printBoolean)
        System.out.println("addVariable() adding variable " +variable);
      this.variableMap.hashMap.put
        (variable, new VariableOccurence
          (intV.getOccurences() + 1, intV.getDeclarated()));
    }
  }


  /**
   * Add declared variable to the hashmap
   * If member variable already is in the hashmap then create a 
   * new arraylist entry
   * @param variable <code>String</code> which we add to the hashmap
   * @param variableType <code>VariableType</code> holds the information of 
   * variable's type
   * @param strictOpt <code>boolean</code> holds the information has the
   * no_strict option (check also public, protected or private protected) given
   */
  public void addDeclaredVariable
  (final String variable, final VariableType variableType,
   final boolean strictOpt) {
    VariableOccurence vo;
    
    if (isEmpty(variable))
      return;

    int lineNumber = variableType.getLine(); 
    
    // If local variable not exists declare it
    if (variableType.isMethodParameter() ||
        (!isMemberVariable && 
         ((vo = (VariableOccurence)this.variableMap.hashMap.get
          (variable)) == null))) {
      if (printBoolean)
        System.out.println("Adding declared local variable:" +variable);
      this.variableMap.hashMap.put
        (variable, new VariableOccurence(0, variableType));
    }
    // If member variable
    else if (isMemberVariable) {
      boolean isPrivate = variableType.isPrivate();
      
      if (printBoolean) {
        System.out.println("Adding declared member variable:" +variable+
                           " strictOpt: " +strictOpt+" isPrivate: "
                           +isPrivate);
      }
      
      if (((strictOpt && isPrivate && !variableType.isStatic()))
          || !strictOpt) {
      
        vo = (VariableOccurence)this.variableMap.hashMap.get
	  (variable);
        if (vo != null) {
	  if (printBoolean) 
            System.out.println
              ("Member variable "+variable+" already referenced");
	  this.variableMap.hashMap.put
            (variable, new VariableOccurence(vo.getOccurences()+1, 
                                             variableType));
        }
        else {
	  if (printBoolean) 
            System.out.println
              ("Member variable "+variable+" not declared yet");
          this.variableMap.hashMap.put
            (variable, new VariableOccurence
              (0, variableType));
        }
      }
    }
    if (printBoolean) printArrayList();
  }

  /**
   * Add declared variable to the hashmap
   * If member variable already is in the hashmap then create a 
   * new arraylist entry
   * @param variable <code>String</code> which we add to the hashmap
   * @param lineNumber <code>int</code> line number in where the modifier
   * occured for the first time
   */
  public void addDeclaredVariable
  (final String variable, final int lineNumber) {
    VariableOccurence vo;
    
    if (isEmpty(variable))
      return;

    // If local variable does not exists then declare it
    if (!isMemberVariable && 
        ((vo = (VariableOccurence)this.variableMap.hashMap.get
          (variable)) == null)) {
      if (printBoolean)
        System.out.println("Adding declared local variable:" +variable);
      this.variableMap.hashMap.put
        (variable, new VariableOccurence(0, new VariableType(lineNumber)));
    }
    // If member variable
    else if (isMemberVariable) {
      if (printBoolean)
        System.out.println("Adding declared member variable:" +variable);

      vo = (VariableOccurence)this.variableMap.hashMap.get
        (variable);
      if (vo != null) {
        if (printBoolean) 
          System.out.println
            ("Member variable "+variable+" already referenced");
        this.variableMap.hashMap.put
          (variable, new VariableOccurence(vo.getOccurences()+1, 
                                           new VariableType(lineNumber)));
      }
      else {
        if (printBoolean) 
          System.out.println
            ("Member variable "+variable+" not declared yet");
        this.variableMap.hashMap.put
          (variable, new VariableOccurence
            (0, new VariableType(lineNumber)));
      }
    }
    if (printBoolean) printArrayList();
  }

  /**
   * Add declared variable to the hashmap
   * If the variable already is in the hashmap then do nothing
   * @param variable <code>String</code> which we add to the hashmap
   */
  public void addDeclaredVariable(final String variable) {
    if (isEmpty(variable))
      return;

    if (printBoolean)
      System.out.println("Adding declared variable:" +variable);
    this.variableMap.hashMap.put(variable, new VariableOccurence(0, true));
  }

  /**
   * Get variable occurence object from the hashmap
   * @param variable <code>String</code> which we get from the hashmap
   * @return variable occurence object
   */
  private VariableOccurence get(final String variable) {
    return (VariableOccurence)this.variableMap.hashMap.get(variable);
  }

  /** Clear a hashmap
   */
  public void clear() {
    variableMap.hashMap.clear();
  }

  /** Get a key (for example key=value) from the string
   * @param entry <code>Object</code> for storing object
   * which is in form KEY=Object
   * @return a key
   */
  public static String getKeyFromEntry(final Object entry) {
    String s = entry.toString();
    
    StringTokenizer st = new StringTokenizer(s == null ? "" : s, "=");

    if (st.hasMoreTokens()) {
      return st.nextToken();
    }
    else {
      return "";
    }
  }

  /** Add hashmap (variables) to the arraylist
   * @param className a class name
   */
  public void addVariables(final String className) {
    if (printBoolean) {
      System.out.println
        ("Adding variables by className:" +className);
      System.out.println
        ("lineNumber with className is: " +
         this.variableMap.getLineNumber());
    }


    if (!isMemberVariable) {
      if (arrayListIndex < 0 || tempArrayList.size() < 1)
        return;
      checkWasThereAnyReferencesBefore();
      this.arrayListIndex--;

      this.variableMap.clear();
      this.variableMap = (VariableMap)this.tempArrayList.get
        (this.arrayListIndex);
      this.tempArrayList.remove(this.arrayListIndex);
      if (printBoolean) {
        System.out.println("TempList after remove"); 
        printArrayList(tempArrayList);
        printArrayList();
        System.out.println
          ("addVariables("+className+"): arrayListIndex: " 
           +this.arrayListIndex);
        System.out.println("Currentmap");
        printHashMap();
      }
      return;
    }

    this.variableMap.setClassName(className);
    this.variableMap.setClassStopped(true);
    checkWasThereAnyReferencesBefore();
    this.arrayListIndex--;

    this.arrayList.add((VariableMap)this.variableMap.clone());
    this.variableMap.clear();
    this.variableMap = (VariableMap)this.tempArrayList.get
      (this.arrayListIndex);
    this.tempArrayList.remove(this.arrayListIndex);
    if (printBoolean) {
      System.out.println("TempList after remove"); 
      printArrayList(tempArrayList);
      printArrayList();
      System.out.println
        ("addVariables("+className+"): arrayListIndex: " 
         +this.arrayListIndex);
      System.out.println("Currentmap");
      printHashMap();
    }
  }

  /** Check if the variable exists in the temporary array list
   * and store it if it exists
   * @param variable which should be checked
   */
  private void isModifierInTemporaryList(final String variable) {
    if (printBoolean) {
      System.out.println("isModifierInTemporaryList() begins");
      printArrayList(tempArrayList);
    }
	
    int size = this.arrayListIndex;
    if (size <= this.tempArrayList.size() && size >= 0)
      for (int i = size; i-- != 0;) {
        VariableMap vm = (VariableMap)this.tempArrayList.get(i);
        if (!vm.hasClassStopped) {
          VariableOccurence vo;
          if ((vo = (VariableOccurence)vm.hashMap.get
               (variable)) != null) {
            if (printBoolean) 
              System.out.println
                ("isModifierInTemporaryList(), variable found: "
                 +variable);
            vm.hashMap.put(variable, new VariableOccurence
              (vo.getOccurences() + 1, vo.getDeclarated()));
            variableMap.hashMap.remove(variable);
            break;
          } else {
            if (printBoolean) 
              System.out.println
			    
                ("isModifierInTemporaryList(), variable not found: "    +variable);
            vm.hashMap.put(variable, new VariableOccurence
              (0, false));
            variableMap.hashMap.remove(variable);
            break;
          }
        }
      }

    if (printBoolean) {
      System.out.println("isModifierInTemporaryList() ends");
      printArrayList(tempArrayList);
    }

  }

  /* Check is there any references
   */
  private void checkWasThereAnyReferencesBefore() {
    Set entry = variableMap.hashMap.entrySet();
    Object[] tableEntry = entry.toArray();
	
    for (int j = tableEntry.length; --j >= 0;) {
      String key = getKeyFromEntry(tableEntry[j]);
      VariableOccurence vo = 
        (VariableOccurence)variableMap.hashMap.get(key);
	    
      if (!vo.getDeclarated()) {
        isModifierInTemporaryList(key);
      }
    }
  }

  /* Add hashmap (variables) to the arraylist
   * @param lineNumber and set linenumber for class
   */
  public void addVariables(int lineNumber) {
    if (!isMemberVariable && this.variableMap.hashMap.size() == 0)
      return;

    if (printBoolean) {
      System.out.println("Adding variables to templist, hashmap:");
      System.out.println("isMemberVariable: " +isMemberVariable);
      printHashMap();
    }

    this.variableMap.setLineNumber(oldLineNumber);
    this.tempArrayList.add(this.variableMap.clone());
    this.variableMap.clear();

    this.arrayListIndex++;

    this.variableMap.setLineNumber(lineNumber);
    oldLineNumber = lineNumber;
  }

  /** Add hashmap (variables) to the arraylist
   */
  public void addVariables() {
    if (printBoolean) {
      System.out.println("Adding variables, hashmap:");
      System.out.println("isMemberVariable: " +isMemberVariable);
      printHashMap();
    }

    this.variableMap.setClassStopped(true);
    this.arrayList.add(this.variableMap.clone());
    this.variableMap.clear(); 
  }

  /** Print information
   */
  public void printDebug() {
    int size = arrayList.size();

    printHashMap(variableMap.hashMap);

    for (int i = size; --i >= 0;) {
      System.out.println("\n----------------printDebug----------------\n");
      HashMap hh = ((VariableMap)arrayList.get(i)).hashMap;
      Set entry = hh.entrySet();
      Object[] tableEntry = entry.toArray();
      VariableOccurence vo;
      String name;
      
      for (int j = tableEntry.length; --j >= 0;) {
        name = getKeyFromEntry(tableEntry[j]);
        vo = (VariableOccurence)hh.get(name);
        System.out.println("layer : " +i + " " +name 
			   +" " +vo.getOccurences());
      }
    }
    System.out.println("\n----------------printDebug----------------\n");
  }

  /** Print warning messages
   * @param fileName <code>String</code> holds the Java class file name 
   */
  public void print(final String fileName) {
    String key;
    boolean found;
    String name;
    VariableOccurence vo;

    for (int size = this.arrayList.size(), tempLayer = size;
	 --tempLayer >= 0; ) {

      found = false;
      if (printBoolean) {
        System.out.println("Checking in layer: " +
                           tempLayer + " size " +size);
        printHashMap(((VariableMap)this.arrayList.get
                      (tempLayer)).hashMap);
        printDebug();
      }
    
      HashMap hh = ((VariableMap)arrayList.get(tempLayer)).hashMap;
      if (hh == null)
        return;

      Object[] table = (Object[])hh.values().toArray();
      if (table == null)
        return;
      Set entry = hh.entrySet();
      Object[] tableEntry = entry.toArray();
      if (tableEntry == null)
        return;
   
      for (int j = table.length; --j >= 0;) {
        name = getKeyFromEntry(tableEntry[j]);
        vo = (VariableOccurence)hh.get(name);
        if (printBoolean)
          System.out.println("checking variable : " +
			     name +" " +vo.getOccurences());

        VariableOccurence intId = (VariableOccurence)table[j];
        if (intId != null && intId.getOccurences() == 0 
	    && intId.getDeclarated()) {
          VariableType vt = intId.getVariableType();
          StringBuffer sb;

          if (vt.isMethodParameter())
            sb = new StringBuffer(METHOD_VARIABLE);
          else
            sb = new StringBuffer(VARIABLE); 
          if (vt.isStatic() && isMemberVariable)
            sb.append((key = getKeyFromEntry
                       (tableEntry[j]))).append(" might not be used");
          else if (!vt.isPrivate() && isMemberVariable)
            sb.append((key = getKeyFromEntry
                       (tableEntry[j]))).append(" might not be used");
          else
            sb.append((key = getKeyFromEntry
                       (tableEntry[j]))).append(" is not used");
          int line = intId.getLineNumber();
          
	  System.err.println(fileName+":"+line+": "+sb.toString());
        }
        
      }
    }
  }

  /** Inner class for modifiers
   */
  static final class VariableMap implements Cloneable {
    /* hashMap holds the information of variables */
    public HashMap hashMap;
    /* class name */
    private String className;
    /* line number for class */
    private int lineNumber;
      
    /* boolean flag for declaring if the class has already stopped */
    private boolean hasClassStopped;

    /* Default constructor */
    public VariableMap() {
      hashMap = new HashMap(); className = null;
      lineNumber = -1; hasClassStopped = false;
    }

    /* Clone the object */ 
    public Object clone() {
      VariableMap vm = new VariableMap();
      vm.hashMap = (HashMap)hashMap.clone();
      vm.className = className;
      vm.lineNumber = lineNumber;
      vm.hasClassStopped = hasClassStopped;
      return vm;
    }

    /* Return line number */
    public int getLineNumber() { return lineNumber; }

    /* Set line number */
    public void setLineNumber(final int lineNumber) {
      this.lineNumber = lineNumber; 
    }

    /* Set class stopped flag */
    public void setClassStopped(final boolean hasClassStopped) {
      this.hasClassStopped = hasClassStopped;
    }

    /* Get class stopped flag */
    public boolean hasClassStopped() {
      return this.hasClassStopped;
    }

    /* Clear information of the object */
    public void clear() {
      hashMap.clear();
      className = null; 
      hasClassStopped = false;
    }

    /* Set class name information */
    public void setClassName(final String className) {
      this.className = className; 
    }
    /* Get class name information */
    public String getClassName() { return className; }
  }
  
  /** Static inner class for occurences
   * used for modifier occurences
   */
  static final class VariableOccurence {
    // Variable type object
    private VariableType variableType;
    // Modifier's number of occurences
    private int occurences;
    // Has variable declareted
    private final boolean declarated;
    
    /* Constructor */
    public VariableOccurence(final int occurences) {
      this.occurences = occurences;
      this.declarated = false;
    }

    /* Constructor */
    public VariableOccurence(final int occurences,
                             final VariableType variableType) {
        
      this.occurences = occurences;
      this.variableType = variableType;
      this.declarated = true;
    }

    /* Constructor */
    public VariableOccurence(final int occurences,
                             final boolean declaration) {
      this.occurences = occurences;
      this.declarated = declaration;
    }

    /* Constructor */
    public VariableOccurence(final int occurences,
                             final boolean declaration,
                             final VariableType variableType) {
      this.occurences = occurences;
      this.declarated = declaration;
      this.variableType = variableType;
    }

    /* Return information if the variable has already been declared */
    public boolean getDeclarated() { return declarated; }

    /* Return variable's line number */
    public int getLineNumber() { return variableType != null ? variableType.getLine() : 0; }
    /* Return the number of variable occurences */
    public int getOccurences() { return occurences; }
    /* Return variable's type */
    public VariableType getVariableType() { return variableType; }
    
  }
}
